/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file bt_acl_pkt_buf.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief API for dealing with ACL packet buffers.
 * 
 * ACL packet buffers are used on the receiving side to buffer
 * (connection-less) data packets. Each ACL packet buffer enacapsulates
 * a single ACL packet.
 * 
 * This file provides the Definition of an ACL packet buffer, as well as
 * some usefull functions helping to deal with ACL packet buffers.
 * 
 * Note: Applications typically make not use of the functions defined
 * here. Instead, this API is intended for lower layer protocols.
 */
#ifndef BT_ACL_PKT_BUF_H_
#define BT_ACL_PKT_BUF_H_

#include <sys/types.h>
#include <bt/bt_defs.h>

/**
 * \brief ACL Packet Buffer Definition
 */
struct bt_acl_pkt_buf_s {
	struct bt_hci_pkt_acl* pkt;
						///< Pointer to the received data
	struct bt_acl_pkt_buf_s* next;
						///< Pointer to a next packet buffer
	struct bt_acl_pkt_buf_s* next_frag;
						///< Pointer to the next fragment of the message
	u_char* pload;		///< Pointer to the payload of the received data
	u_short plen;		///< Payload size of the received data
	u_long tstamp;		///< Timestamp, created at packet reception
	void* arg;			///< Pointer to some argument
};

/**
 * \brief ACL packet buffer typedef
 */
typedef struct bt_acl_pkt_buf_s bt_acl_pkt_buf;


/**
 * \brief Typedef of a FIFO queue, buffering ACL packet buffers
 */
typedef struct bt_acl_pkt_fifo_s bt_acl_pkt_fifo;


/**
 * \name ACL Packet Buffer Creation & Destruction
 * 
 * Definition of an ACL packet buffer and API functions for
 * creating / destroying ACL packet buffers
 */
//@{

/**
 * \brief Creates a single ACL packet buffer.
 * 
 * Note: The memory allocated for an ACL packet buffer is determined
 * by the ACL packet types in use. This has to be defined during
 * initialization. See #bt_acl_init for further details.
 * 
 * \returns
 * - Pointer to the created buffer
 * - 0 if not enough memory available
 * 
 * \see 
 * #bt_acl_init \n
 * #bt_acl_pkt_buf \n
 * #bt_acl_pkt_buf_destroy
 */
bt_acl_pkt_buf* bt_acl_pkt_buf_create(struct btstack* stack);

/**
 * \brief Destroys the ACL packet buffer passed.
 * 
 * Destroys a previousely created (using #bt_acl_pkt_buf_create)
 * ACL packet buffer.
 * 
 * \see
 * #bt_acl_pkt_buf \n
 * #bt_acl_pkt_buf_create
 */
void bt_acl_pkt_buf_destroy(bt_acl_pkt_buf* buf);

/**
 * \brief Creates a queue of ACL packet buffers
 * 
 * Creates a queue (i.e. a linked list) of ACL packet buffers
 * of the size given. 
 * 
 * Note: The memory allocated for a single ACL packet buffer is determined
 * by the ACL packet types in use. This has to be defined during
 * initialization (see #bt_acl_init).
 * 
 * \param nr_bufs Number of buffers to create
 * 
 * \returns
 * - Pointer to the created queue
 * - 0 if not enough memory available
 * 
 * \see
 * #bt_acl_pkt_buf \n
 * #bt_acl_pkt_queue_destroy
 */
bt_acl_pkt_buf* bt_acl_pkt_queue_create(struct btstack* stack, u_short nr_bufs);

/**
 * \brief Destroys an ACL packet buffer queue
 * 
 * Destroys the previousely created (using #bt_acl_pkt_queue_create) queue
 * of ACL packet buffers.
 * 
 * \param que Pointer to the first packet buf of the queue
 * 
 * \see
 * #bt_acl_pkt_buf \n
 * #bt_acl_pkt_queue_create
 */
void bt_acl_pkt_queue_destroy(bt_acl_pkt_buf* que);
//@}


/**
 * \name FIFO queueing ACL packet buffers
 */
//@{

/**
 * \brief Creates a fifo queue for queueing ACL packet buffers.
 * 
 * \returns
 * - Pointer to the created packet queue
 * - 0 if not enough memory available
 * 
 * \see
 * #bt_acl_pkt_fifo_add_first \n
 * #bt_acl_pkt_fifo_add_last \n
 * #bt_acl_pkt_fifo_get_first \n
 * #bt_acl_pkt_fifo_destroy
 */
bt_acl_pkt_fifo* bt_acl_pkt_fifo_create(void);

/**
 * \brief Destroys a FIFO queue.
 * 
 * Destroys a previousely created ACL packet buffer FIFO queue.
 * 
 * \param fifo Pointer to the fifo queue to destroy
 * 
 * \see
 * #bt_acl_pkt_fifo_create \n
 * #bt_acl_pkt_fifo_add_first \n
 * #bt_acl_pkt_fifo_add_last \n
 * #bt_acl_pkt_fifo_get_first
 */
void bt_acl_pkt_fifo_destroy(bt_acl_pkt_fifo* fifo);

/**
 * \brief Gets the first ACL packet buffer from the FIFO queue passed.
 * 
 * Removes the first ACL packet buffer from the FIFO queue passed
 * and returns a pointer to it.
 * 
 * \param fifo Pointer to the fifo queue
 * 
 * \returns Pointer to the first ACL packet buffer, or NULL if the
 * FIFO queue is empty.
 * 
 * \see
 * #bt_acl_pkt_fifo_create \n
 * #bt_acl_pkt_fifo_destroy \n
 * #bt_acl_pkt_fifo_add_first \n
 * #bt_acl_pkt_fifo_add_last
 */
bt_acl_pkt_buf* bt_acl_pkt_fifo_get_first(bt_acl_pkt_fifo* fifo);

/**
 * \brief Returns a pointer to the first ACL packet buffer in the
 * queue passed.
 * 
 * Note: the ACL packet buffer won't be removed from the queue.
 * 
 * \param fifo Pointer to the fifo queue
 * 
 * \returns Pointer to the first ACL packet buffer, or NULL if the
 * FIFO queue is empty.
 * 
 * \see
 * #bt_acl_pkt_fifo_add_first \n
 * #bt_acl_pkt_fifo_add_last \n
 * #bt_acl_pkt_fifo_get_first \n
 * #bt_acl_pkt_fifo_destroy
 */
bt_acl_pkt_buf* bt_acl_pkt_fifo_top(bt_acl_pkt_fifo* fifo);

/**
 * \brief Adds the ACL packet buffer passed to the head of
 * the FIFO queue specified.
 * 
 * \param fifo Pointer to the fifo queue the buffer shall be added to
 * \param buf Pointer to the ACL packet buffer to be added
 *
 * \see
 * #bt_acl_pkt_fifo_create \n
 * #bt_acl_pkt_fifo_destroy \n
 * #bt_acl_pkt_fifo_add_last \n
 * #bt_acl_pkt_fifo_get_first
 */
void bt_acl_pkt_fifo_add_first(bt_acl_pkt_fifo* fifo, bt_acl_pkt_buf* buf);

/**
 * \brief Adds the ACL packet buffer passed to the end of the
 * FIFO queue specified.
 * 
 * \param fifo Pointer to the fifo queue the buffer shall be added to
 * \param buf Pointer to the ACL packet buffer to be added
 * 
 * #bt_acl_pkt_fifo_create \n
 * #bt_acl_pkt_fifo_destroy \n
 * #bt_acl_pkt_fifo_add_first \n
 * #bt_acl_pkt_fifo_get_first
 */
void bt_acl_pkt_fifo_add_last(bt_acl_pkt_fifo* fifo, bt_acl_pkt_buf* buf);
//@}

#endif /*BT_ACL_PKT_BUF_H_*/
