/*
 * Copyright (C) 2005 Mario Strasser <mast@gmx.net>,
 *                    Swiss Federal Institute of Technology (ETH) Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ccc.h,v 1.1 2005/08/15 10:01:44 mast2 Exp $
 */

/**
 * @file ccc.h
 * @author Mario Strasser <mast@gmx.net>
 * @brief Packet-communication library for the CC1000 radio.
 */
 
#ifndef _CCOM_H_
#define _CCOM_H_
 
#include <sys/types.h>
#include <cc/mac.h>

/**
 * Structure of a CC1000 communication (ccc) packet.
 */
typedef struct ccc_packet_st {
    /** source of the packet */
    u_short src;
    /** destination of the packet */
    u_short dst; 
    /** payload length */
    u_short length;
    /** packet type */
    u_char type;
    /** payload data */
    u_char data[0];
} __attribute__((__packed__)) ccc_packet_t;
 
/** packet-handler type */
typedef void (*ccc_packet_handler_t)(ccc_packet_t *pkt);
 
/**
 * Initialises the CC1000 communication library. 
 *
 * @param mac_interface [in] CC1000 MAC-Interface to use 
 * @return 0 on success, -1 otherwise.
 */
int ccc_init(mac_interface_t *mac_interface);

/**
 * Releases the CC1000 communication library.
 */ 
void ccc_release(void);

/**
 * Returns the configured mac address.
 * 
 * @return configured mac address.
 */
u_short ccc_get_address(void);

/**
 * Returns the maximal payload size of a ccc packet.
 * 
 * @return maximal payload size of a ccc packet.
 */
u_short ccc_get_mtu(void);

/**
 * Allocates memory for a new ccc packet with the given (maximal) payload size.
 * 
 * @param data_length [in] maximal length of the payload
 * @return pointer to the new ccc packet.
 */
ccc_packet_t* new_ccc_packet(u_short data_length);

/**
 * Sends a ccc packet to the specified destination node. 
 * 
 * @param dst [in] destination address
 * @param type [in] packet type
 * @param pkt [in] packet to send, note that the header will be overwritten
 * @return 0 on success, -1 otherwise.
 */
int ccc_send(u_short dst, u_char type, ccc_packet_t *pkt);
 
/**
 * Registers a packet handler for the specified packet type.
 * 
 * @param type [in] packet type to handle
 * @param handler [in] packet handler to use
 * @returns 0 on success, -1 if the specified type is invalid or already
 *          assigned to an other handler.
 */
int ccc_register_packet_handler(u_char type, ccc_packet_handler_t handler);

/**
 * Deregisters the packet handler of the specified type.
 * 
 * @param type [in] packet type
 */ 
void ccc_deregister_packet_handler(u_char type);

/**
 * Returns the currently installed packet handler for the specified type. 
 * 
 * @param type [in] packet type
 * @return currently installed packet handler, or NULL if there is none.
 */ 
ccc_packet_handler_t ccc_get_packet_handler(u_char type);

void ccc_test(void);

#endif /* _CCOM_H_ */
