/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

 /*
 * btn-terminal.h
 *
 * 11.06.2004 Martin Hinz <btnode@hinz.ch>
 */
#ifndef TERMINAL_H_INCLUDED
#define TERMINAL_H_INCLUDED


#include <stdio.h>
#include <io.h>

/**

\file btn-terminal.h
\author Martin Hinz <btnode@hinz.ch>

\brief Easy use for user input/output and execute commands.

There are some predefined commands to control nut-os, btnode and bluetooth.

@see 
\link bt-cmds.h bt-cmds.h\endlink \n
\link acl-com-cmds.h acl-com-cmds.h\endlink \n
\link nut-cmds.h nut-cmds.h\endlink \n
\link btn-cmds.h btn-cmds.h\endlink \n

*/

/*!
 * Maximal cmds that can be registered. This number can be ajusted.
 */
#define MAX_CMDS 50

/*!
 * Maximal length of terminal input. This number can be ajusted.
 */
#define INPUT_LENGTH 130

/*!
 * Maximal history depth. This number can be ajusted.
 */
#define HISTORY_DEPTH 5

/*!
 * Maximal length of command name. This number can be ajusted.
 */
#define LEN_CMD_NAME 10

/*!
 * Passing argument to start (fork) a new terminal thread.
 */
#define BTN_TERMINAL_FORK 0

/*!
 * Passing argument to use main thread for terminal input (no new thread).
 */
#define BTN_TERMINAL_NOFORK 1

/*!
 * Maximum length of the terminal prompt
 */
#define TERMINAL_MAX_PROMPT_LENGTH 20

#define TERMINAL_THREADBUFFER_SIZE           1023
#define TERMINAL_MAX_LOGGED_THREADS          2

/*! \name DEBUG with mutex support provided in each DEBUG call
 * following function write with mutex to the debug uart!\n
 * Advantage: No output-mix due to different threads writing to same uart.\n
 * Disadvantage: Still mix of single DEBUG calls of different threads.
 */
//@{
/*!
 * to print a buffer string ending with '\\0'.\n
 * This call uses mutex to access the uart!
 */
void DEBUGB(char* text);

/*!
 * internal function used by #DEBUGT. \n 
 */
int DEBUGT_I(PGM_P fmt, ...);

/*!
 * to print a formated string with arguments or just plain text
 * use only #DEBUGT to put the text into ROM memory! \n
 * The DEBUGT_I() is an internal function.\n
 * This call uses mutex to access the uart!
 */
#define DEBUGT(text, ...) DEBUGT_I(PSTR(text),## __VA_ARGS__)

/*!
 * to print buffer with length len.\n
 * This call uses mutex to access the uart!
 */
void DEBUGL(char* buf, u_char len);
//@}

/*! \name DEBUG with mutex support provided by a lock function
 * following function write directly to uart without mutex!\n
 * use btn_terminal_get_writing_lock() for unique debug uart access and
 * btn_terminal_free_writing_lock() to free the lock again\n
 * Advantage: Unique output of one thread also with several DEBUG calls.\n
 * Disadvantage: Lock and unlock the unique uart access.
 */


//@{
/*!
 * To print a buffer string ending with '\\0'.\n
 * This call does not use mutex. Use btn_terminal_get_writing_lock() and
 * btn_terminal_free_writing_lock() if mutex access is demanded
 */
#define DEBUGB_M(text) fputs(text,debug_uart);

/*!
 * To print a formated string with arguments or just plain text.\n
 * This call does not use mutex. Use btn_terminal_get_writing_lock() and
 * btn_terminal_free_writing_lock() if mutex access is demanded
 */
#define DEBUGT_M(text,...) fprintf_P(debug_uart,PSTR(text),## __VA_ARGS__)

/*!
 * to print buffer with length len.\n
 * This call does not use mutex. Use btn_terminal_get_writing_lock() and
 * btn_terminal_free_writing_lock() if mutex access is demanded
 */
#define DEBUGL_M(buf, len) _write(_fileno(debug_uart), buf, len)

/**
 * \brief use this function to capture terminal output.
 * 
 * Prints the formatted string passed to the terminal and captures the
 * printed string using dedicated thread buffers.
 * 
 * Note: The terminal output is also written to the log buffer using class 1,
 * log level #LOG_INFO. Thus, to enable terminal output logging, call either
 * #log_set_logmask or use the corresponding terminal command.
 * 
 * \see #term_register_thread
 * \see #term_start_capture
 * \see #term_stop_capture
 * \see #log_cmds_init
 */
int tprintf(CONST char* text, ...);

/*******************************************************************************
 * term_register_thread
 ******************************************************************************/
/**
 * @brief Register the calling thread.
 * 
 * Register the calling thread.
 * 
 * @return 0 on success, 1 otherwise.
 */
char term_register_thread(void);

/*******************************************************************************
 * term_start_capture
 ******************************************************************************/
/**
 * @brief Start the recording of terminal output.
 * 
 * Starts the recording of terminal output. After this call, all data written
 * to the terminal using tprintf() is stored in a dedicated thread
 * buffer. To get the buffer of the running thread, use #term_stop_capture().
 * 
 * @return 0 on success, 1 otherwise.
 */
char term_start_capture( void);

/*******************************************************************************
 * term_stop_capture
 ******************************************************************************/
/**
 * @brief Stop the recording of terminal output.
 * 
 * @return Pointer to buffer on success, NULL otherwise.
 */
char *term_stop_capture( void);

/*!
 * use this function to get explicit access to the debug uart
 * and write output with: DEBUGB_M(), DEBUGT_M() and DEBUGL_M()
 * call btn_terminal_free_writing_lock to free lock again
 */
void btn_terminal_get_writing_lock(void);

/*!
 * function to free the explicit access lock on debug uart
 */
void btn_terminal_free_writing_lock(void);

//@}
/*!
 * file pointer to debug device.
 */
extern FILE *debug_uart;

/*!
 * The terminal will read and write to the file pointer passed with the
 * uart argument. An initial prompt can be set. To change the prompt use
 * the function: btn_terminal_set_prompt().
 * @param uart Pointer to device for in- and output
 * @param a_prompt Prompt to be displayed
 */
void btn_terminal_init(FILE *uart, char *a_prompt);

/*!
 * To start the terminal use this function.\n
 * Pass the type #BTN_TERMINAL_FORK to start an own terminal thread
 * or use #BTN_TERMINAL_NOFORK to use the main thread for the terminal thread
 * (in the later case, function call will never return)\n
 * pass thread_size in case of #BTN_TERMINAL_FORK
 * @param type #BTN_TERMINAL_FORK or #BTN_TERMINAL_NOFORK
 * @param thread_size In case of #BTN_TERMINAL_FORK the size of the terminal thread.
 */
void btn_terminal_run(u_char type, u_short thread_size);

/*!
 * register the cmds with a command name and a function pointer. the
 * function will be called with a char pointer argument passing the
 * arguments AFTER the command name.
 * @param name Name of the command
 * @param fkt Pointer to the command function
 */
void btn_terminal_register_cmd(char* name, void (*fkt) (char*));

/*!
 * This function provides the processing of a cmd not typed in
 * by the terminal.\n
 * Make sure cmd is terminated by '\\0'
 * @param cmd The command string same as typed in with terminal
 * @return 0 on success, 1 if cmd was not found
 */
u_char btn_terminal_process_cmd(char* cmd);

/*!
 * to refresh the command line
 */
void btn_terminal_refresh(void);

/*!
 * to set a new prompt
 * @param a_prompt New prompt to be displayed.
 */
void btn_terminal_set_prompt(char *a_prompt);

void btn_set_error(u_char errorno, char* lasterror);

void btn_show_error(void);

#endif
